# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import httpx
import numpy as np
import pytest
import pytest_asyncio

from kserve import InferenceRESTClient, InferRequest, InferInput
from kserve.model_server import app as kserve_app
from kserve.errors import UnsupportedProtocol
from kserve.inference_client import RESTConfig
from kserve.protocol.infer_type import RequestedOutput
from test.test_server import DummyModel


@pytest.mark.asyncio
class TestInferenceRESTClient:

    @pytest_asyncio.fixture(scope="class")
    async def app(self, server):
        model = DummyModel("TestModel")
        model.load()
        not_ready_model = DummyModel("NotReadyModel")
        # model.load()  # Model not loaded, i.e. not ready
        server.register_model(model)
        server.register_model(not_ready_model)
        yield kserve_app
        await server.model_repository_extension.unload("TestModel")
        await server.model_repository_extension.unload("NotReadyModel")

    @pytest_asyncio.fixture(scope="class")
    async def rest_client(self, request, app):
        config = RESTConfig(
            transport=httpx.ASGITransport(app=app), verbose=True, protocol=request.param
        )
        client = InferenceRESTClient(config=config)
        yield client
        await client.close()

    @pytest.mark.parametrize(
        "rest_client, protocol", [("v1", "v1"), ("v2", "v2")], indirect=["rest_client"]
    )
    async def test_is_server_ready(self, rest_client, protocol):
        if protocol == "v2":
            assert (
                await rest_client.is_server_ready(
                    "http://test-server",
                    headers={"Host": "test-server.com"},
                    timeout=1.3,
                )
                is True
            )
        else:
            # Unsupported protocol
            with pytest.raises(UnsupportedProtocol, match="Unsupported protocol v1"):
                await rest_client.is_server_ready(
                    "http://test-server",
                    headers={"Host": "test-server.com"},
                    timeout=1.3,
                )

    @pytest.mark.parametrize(
        "rest_client, protocol",
        [("v1", "v1"), ("v2", "v2"), ("v3", "v3")],
        indirect=["rest_client"],
    )
    async def test_is_server_live(self, rest_client, protocol):
        if protocol == "v3":
            # Unsupported protocol
            with pytest.raises(UnsupportedProtocol, match="Unsupported protocol v3"):
                await rest_client.is_server_live(
                    "http://test-server",
                    headers={"Host": "test-server.com"},
                    timeout=1.3,
                )
        else:
            assert (
                await rest_client.is_server_live(
                    "http://test-server",
                    headers={"Host": "test-server.com"},
                    timeout=1.3,
                )
                is True
            )

    @pytest.mark.parametrize(
        "rest_client, protocol",
        [("v1", "v1"), ("v2", "v2"), ("v3", "v3")],
        indirect=["rest_client"],
    )
    async def test_is_model_ready(self, rest_client, protocol):
        if protocol == "v3":
            # Unsupported protocol
            with pytest.raises(UnsupportedProtocol, match="Unsupported protocol v3"):
                await rest_client.is_model_ready(
                    "http://test-server",
                    "TestModel",
                    headers={"Host": "test-server.com"},
                    timeout=1.3,
                )
        else:

            # Ready model
            assert (
                await rest_client.is_model_ready(
                    "http://test-server/",
                    "TestModel",
                    headers={"Host": "test-server.com"},
                    timeout=1.3,
                )
                is True
            )
            # Not ready model
            assert (
                await rest_client.is_model_ready(
                    "http://test-server",
                    "NotReadyModel",
                    headers={"Host": "test-server.com"},
                    timeout=1.3,
                )
                is False
            )

    @pytest.mark.parametrize(
        "rest_client, protocol",
        [("v1", "v1"), ("v2", "v2"), ("v3", "v3")],
        indirect=["rest_client"],
    )
    async def test_infer(self, rest_client, protocol):
        if protocol == "v1":
            input_data = {"instances": [1, 2]}
            res = await rest_client.infer(
                "http://test-server/",
                model_name="TestModel",
                data=input_data,
                headers={"Host": "test-server.com"},
                timeout=2,
            )
            assert res["predictions"] == [1, 2]

            input_data = {"inputs": [1, 2]}
            res = await rest_client.infer(
                "http://test-server/",
                model_name="TestModel",
                data=input_data,
                headers={"Host": "test-server.com"},
                timeout=2,
            )
            assert res["predictions"] == [1, 2]

        elif protocol == "v2":
            request_id = "2ja0ls9j1309"
            input_data = {
                "id": request_id,
                "parameters": {"test-param": "abc"},
                "inputs": [
                    {
                        "name": "input-0",
                        "datatype": "INT32",
                        "shape": [2, 2],
                        "data": [[1, 2], [3, 4]],
                        "parameters": {"test-param": "abc"},
                    }
                ],
            }
            res = await rest_client.infer(
                "http://test-server/",
                model_name="TestModel",
                data=input_data,
                headers={"Host": "test-server.com"},
                timeout=2,
            )
            assert res.outputs[0].data == [1, 2, 3, 4]
            assert res.id == request_id

            input_data = InferRequest(
                model_name="TestModel",
                request_id=request_id,
                infer_inputs=[
                    InferInput(
                        name="input-0",
                        datatype="INT32",
                        shape=[2, 2],
                        data=[[1, 2], [3, 4]],
                        parameters={"test-param": "abc"},
                    )
                ],
                parameters={"test-param": "abc"},
            )

            res = await rest_client.infer(
                "http://test-server/",
                model_name="TestModel",
                data=input_data,
                headers={"Host": "test-server.com"},
                timeout=2,
            )
            assert res.outputs[0].data == [1, 2, 3, 4]
            assert res.id == request_id
        else:
            # Unsupported protocol
            input_data = {"instances": [1, 2]}
            with pytest.raises(UnsupportedProtocol, match="Unsupported protocol v3"):
                await rest_client.infer(
                    "http://test-server/",
                    model_name="TestModel",
                    data=input_data,
                    headers={"Host": "test-server.com"},
                    timeout=2,
                )

        # model_name not provided
        input_data = {"instances": [1, 2]}
        with pytest.raises(ValueError):
            await rest_client.infer(
                "http://test-server/",
                data=input_data,
                headers={"Host": "test-server.com"},
                timeout=2,
            )

    @pytest.mark.parametrize(
        "rest_client, protocol",
        [("v2", "v2")],
        indirect=["rest_client"],
    )
    async def test_infer_with_all_binary_data(self, rest_client, protocol):
        request_id = "2ja0ls9j1309"
        fp16_data = np.array([[1.1, 2.22], [3.345, 4.34343]], dtype=np.float16)
        input_data = InferRequest(
            model_name="TestModel",
            request_id=request_id,
            infer_inputs=[
                InferInput(
                    name="input-0",
                    datatype="INT32",
                    shape=[2, 2],
                    data=[[1, 2], [3, 4]],
                    parameters={"test-param": "abc"},
                ),
                InferInput(
                    name="fp16_data",
                    datatype="FP16",
                    shape=[2, 2],
                    parameters={"test-param": "abc"},
                ),
            ],
            parameters={"test-param": "abc", "binary_data_output": True},
        )
        input_data.inputs[1].set_data_from_numpy(fp16_data, binary_data=True)

        res = await rest_client.infer(
            "http://test-server/",
            model_name="TestModel",
            data=input_data,
            headers={"Host": "test-server.com"},
            timeout=2,
        )
        assert res.outputs[0].data == [1, 2, 3, 4]
        assert res.id == request_id

    @pytest.mark.parametrize(
        "rest_client, protocol",
        [("v2", "v2")],
        indirect=["rest_client"],
    )
    async def test_infer_with_binary_data(self, rest_client, protocol):
        request_id = "2ja0ls9j1309"
        fp16_data = np.array([[1.1, 2.22], [3.345, 4.34343]], dtype=np.float16)
        input_data = InferRequest(
            model_name="TestModel",
            request_id=request_id,
            infer_inputs=[
                InferInput(
                    name="input-0",
                    datatype="INT32",
                    shape=[2, 2],
                    data=[[1, 2], [3, 4]],
                    parameters={"test-param": "abc"},
                ),
                InferInput(
                    name="fp16_data",
                    datatype="FP16",
                    shape=[2, 2],
                    parameters={"test-param": "abc"},
                ),
            ],
            parameters={"test-param": "abc"},
            request_outputs=[
                RequestedOutput(
                    name="output-0", parameters={"binary_data_output": True}
                )
            ],
        )
        input_data.inputs[1].set_data_from_numpy(fp16_data, binary_data=True)

        res = await rest_client.infer(
            "http://test-server/",
            model_name="TestModel",
            data=input_data,
            headers={"Host": "test-server.com"},
            timeout=2,
        )
        assert res.outputs[0].data == [1, 2, 3, 4]
        assert res.id == request_id

    @pytest.mark.parametrize(
        "rest_client", ["v1", "v2", "v3"], indirect=["rest_client"]
    )
    async def test_infer_graph_endpoint(self, rest_client, httpx_mock):
        request_id = "2ja0ls9j1309"
        model_name = "TestModel"
        httpx_mock.add_response(
            url="http://test-server-v1", json={"predictions": [1, 2]}
        )
        httpx_mock.add_response(
            url="http://test-server-v2",
            json={
                "id": request_id,
                "model_name": model_name,
                "outputs": [
                    {
                        "name": "output-0",
                        "datatype": "INT32",
                        "shape": [2, 2],
                        "data": [1, 2, 3, 4],
                    }
                ],
            },
        )

        async with httpx.AsyncClient() as client:
            rest_client._client = client
            input_data = {"instances": [1, 2]}
            res = await rest_client.infer(
                "http://test-server-v1",
                data=input_data,
                headers={"Host": "test-server.com"},
                timeout=2,
                is_graph_endpoint=True,
            )
            assert res["predictions"] == [1, 2]

            input_data = {
                "id": request_id,
                "parameters": {"test-param": "abc"},
                "inputs": [
                    {
                        "name": "input-0",
                        "datatype": "INT32",
                        "shape": [2, 2],
                        "data": [[1, 2], [3, 4]],
                        "parameters": {"test-param": "abc"},
                    }
                ],
            }
            res = await rest_client.infer(
                "http://test-server-v2",
                data=input_data,
                headers={"Host": "test-server.com"},
                timeout=2,
                is_graph_endpoint=True,
            )
            assert res["outputs"][0]["data"] == [1, 2, 3, 4]
            assert res["id"] == request_id

            input_data = InferRequest(
                model_name=model_name,
                request_id=request_id,
                infer_inputs=[
                    InferInput(
                        name="input-0",
                        datatype="INT32",
                        shape=[2, 2],
                        data=[[1, 2], [3, 4]],
                        parameters={"test-param": "abc"},
                    )
                ],
                parameters={"test-param": "abc"},
            )

            res = await rest_client.infer(
                "http://test-server-v2",
                data=input_data,
                headers={"Host": "test-server.com"},
                timeout=2,
                is_graph_endpoint=True,
            )
            assert res["outputs"][0]["data"] == [1, 2, 3, 4]
            assert res["id"] == request_id

    @pytest.mark.parametrize(
        "rest_client, protocol",
        [("v1", "v1"), ("v2", "v2"), ("v3", "v3")],
        indirect=["rest_client"],
    )
    async def test_infer_path_based_routing(self, rest_client, protocol, httpx_mock):
        request_id = "2ja0ls9j1309"
        model_name = "TestModel"
        async with httpx.AsyncClient() as client:
            rest_client._client = client
            if protocol == "v1":
                httpx_mock.add_response(
                    url=re.compile(r"http://test-server/serving/test/test-isvc/v1/*"),
                    json={"predictions": [1, 2]},
                )
                input_data = {"instances": [1, 2]}
                res = await rest_client.infer(
                    "http://test-server/serving/test/test-isvc",
                    model_name="TestModel",
                    data=input_data,
                    headers={"Host": "test-server.com"},
                    timeout=2,
                )
                assert res["predictions"] == [1, 2]
            elif protocol == "v2":
                httpx_mock.add_response(
                    url=re.compile(r"http://test-server/serving/test/test-isvc/v2/*"),
                    json={
                        "id": request_id,
                        "model_name": model_name,
                        "outputs": [
                            {
                                "name": "output-0",
                                "datatype": "INT32",
                                "shape": [2, 2],
                                "data": [1, 2, 3, 4],
                            }
                        ],
                    },
                )
                input_data = {
                    "id": request_id,
                    "parameters": {"test-param": "abc"},
                    "inputs": [
                        {
                            "name": "input-0",
                            "datatype": "INT32",
                            "shape": [2, 2],
                            "data": [[1, 2], [3, 4]],
                            "parameters": {"test-param": "abc"},
                        }
                    ],
                }
                res = await rest_client.infer(
                    "http://test-server/serving/test/test-isvc",
                    model_name="TestModel",
                    data=input_data,
                    headers={"Host": "test-server.com"},
                    timeout=2,
                )
                assert res.outputs[0].data == [1, 2, 3, 4]
                assert res.id == request_id

                input_data = InferRequest(
                    model_name="TestModel",
                    request_id=request_id,
                    infer_inputs=[
                        InferInput(
                            name="input-0",
                            datatype="INT32",
                            shape=[2, 2],
                            data=[[1, 2], [3, 4]],
                            parameters={"test-param": "abc"},
                        )
                    ],
                    parameters={"test-param": "abc"},
                )

                res = await rest_client.infer(
                    "http://test-server/serving/test/test-isvc",
                    model_name=model_name,
                    data=input_data,
                    headers={"Host": "test-server.com"},
                    timeout=2,
                )
                assert res.outputs[0].data == [1, 2, 3, 4]
                assert res.id == request_id
            else:
                # Unsupported protocol
                input_data = {"instances": [1, 2]}
                with pytest.raises(
                    UnsupportedProtocol, match="Unsupported protocol v3"
                ):
                    await rest_client.infer(
                        "http://test-server/serving/test/test-isvc",
                        model_name="TestModel",
                        data=input_data,
                        headers={"Host": "test-server.com"},
                        timeout=2,
                    )

    @pytest.mark.parametrize(
        "rest_client, protocol", [("v1", "v1"), ("v2", "v2")], indirect=["rest_client"]
    )
    async def test_explain(self, rest_client, protocol):
        if protocol == "v1":
            input_data = {"instances": [1, 2]}
            res = await rest_client.explain(
                "http://test-server",
                "TestModel",
                data=input_data,
                headers={"Host": "test-server.com"},
                timeout=2,
            )
            assert res == {"predictions": [1, 2]}

            input_data = {"inputs": [1, 2]}
            res = await rest_client.explain(
                "http://test-server/",
                "TestModel",
                data=input_data,
                headers={"Host": "test-server.com"},
                timeout=2,
            )
            assert res == {"predictions": [1, 2]}
        else:
            # Unsupported protocol
            with pytest.raises(UnsupportedProtocol, match="Unsupported protocol v2"):
                input_data = input_data = {
                    "parameters": {"test-param": "abc"},
                    "inputs": [
                        {
                            "name": "input-0",
                            "datatype": "INT32",
                            "shape": [2, 2],
                            "data": [[1, 2], [3, 4]],
                            "parameters": {"test-param": "abc"},
                        }
                    ],
                }
                await rest_client.explain(
                    "http://test-server",
                    "TestModel",
                    data=input_data,
                    headers={"Host": "test-server.com"},
                    timeout=2,
                )

    @pytest.mark.parametrize("rest_client", ["v2"], indirect=["rest_client"])
    async def test_base_url_with_no_scheme(self, rest_client):
        with pytest.raises(
            httpx.InvalidURL,
            match="Base url should have 'http://' or 'https://' protocol",
        ):
            await rest_client.is_server_ready(
                "test-server.com",
                headers={"Host": "test-server.com"},
                timeout=1.3,
            )



# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from kserve.utils.inference_client_factory import InferenceClientFactory
from kserve.inference_client import InferenceRESTClient, RESTConfig


@pytest.fixture
def factory():
    return InferenceClientFactory()


def test_get_rest_client_v1(factory):
    config = RESTConfig(protocol="v1")
    client = factory.get_rest_client(config)
    assert client._config.protocol == "v1"
    assert factory._rest_v1_client is client


def test_get_rest_cleint_v1_config_none(factory):
    client = factory.get_rest_client()
    assert isinstance(client, InferenceRESTClient)
    assert client._config.protocol == "v1"
    assert factory._rest_v1_client is client


def test_get_rest_client_v2(factory):
    config = RESTConfig(protocol="v2")
    client = factory.get_rest_client(config)
    assert isinstance(client, InferenceRESTClient)
    assert client._config.protocol == "v2"
    assert factory._rest_v2_client is client


def test_get_rest_client_v1_cached(factory):
    config = RESTConfig(protocol="v1")
    client1 = factory.get_rest_client(config)
    client2 = factory.get_rest_client(config)
    assert client1._config.protocol == "v1"
    assert client1 is client2


def test_get_rest_client_v2_cached(factory):
    config = RESTConfig(protocol="v2")
    client1 = factory.get_rest_client(config)
    client2 = factory.get_rest_client(config)
    assert client1._config.protocol == "v2"
    assert client1 is client2



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from vllm.entrypoints.openai.protocol import (
    ChatCompletionResponseChoice as ChatCompletionChoice,
    ChatCompletionLogProb,
    ChatCompletionLogProbs,
    ChatCompletionStreamResponse as ChatCompletionChunk,
    ChatCompletionResponseStreamChoice as ChunkChoice,
    ChatMessage,
    DeltaMessage as ChoiceDelta,
    CompletionResponseChoice as CompletionChoice,
    CompletionStreamResponse as CompletionChunk,
    CompletionResponseStreamChoice as CompletionChunkChoice,
    CompletionLogProbs,
    UsageInfo,
    ChatCompletionLogProbsContent,
    ModelCard as Model,
    ModelList,
)
from vllm.entrypoints.openai.protocol import ChatCompletionRequest, ChatCompletionResponse as ChatCompletion
from vllm.entrypoints.openai.protocol import CompletionRequest, CompletionResponse as Completion
from vllm.entrypoints.openai.protocol import EmbeddingRequest, EmbeddingResponse as Embedding, EmbeddingResponseData, EmbeddingCompletionRequest
from vllm.entrypoints.openai.protocol import RerankRequest, RerankResponse as Rerank
from vllm.entrypoints.chat_utils import (
    ChatCompletionContentPartParam,
    CustomChatCompletionMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionAssistantMessageParam,
    ConversationMessage,
)

from typing import Optional
from pydantic import BaseModel, Field


class Error(BaseModel):
    code: Optional[str] = Field(...)
    message: str
    param: Optional[str] = Field(...)
    type: str


class ErrorResponse(BaseModel):
    error: Error


__all__ = [
    "ChatCompletion",
    "ChatCompletionChoice",
    "ChatCompletionChunk",
    "CompletionChunk",
    "CompletionChunkChoice",
    "ChatMessage",
    "ChatCompletionLogProb",
    "ChatCompletionLogProbs",
    "ChoiceDelta",
    "ChunkChoice",
    "Completion",
    "CompletionChoice",
    "ChatCompletionRequest",
    "CompletionRequest",
    "ChatCompletionContentPartParam",
    "CustomChatCompletionMessageParam",
    "ChatCompletionMessageParam",
    "ChatCompletionToolMessageParam",
    "ChatCompletionContentPartTextParam",
    "ChatCompletionAssistantMessageParam",
    "ConversationMessage",
    "Error",
    "ErrorResponse",
    "CompletionLogProbs",
    "UsageInfo",
    "ChatCompletionLogProbsContent",
    "Embedding",
    "EmbeddingCompletionRequest",
    "EmbeddingRequest",
    "EmbeddingResponseData",
    "Model",
    "ModelList",
    "RerankRequest",
    "Rerank",
]



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import Any, AsyncGenerator, AsyncIterator, Callable, Dict, Optional, Union
import inspect

from pydantic import BaseModel
from fastapi import Request

from kserve.protocol.rest.openai.types import (
    ChatCompletion,
    Completion,
    CompletionRequest,
    ChatCompletionRequest,
    EmbeddingRequest,
    Embedding,
    ErrorResponse,
    RerankRequest,
    Rerank,
)

from ....model import BaseKServeModel


class ChatPrompt(BaseModel):
    response_role: str = "assistant"
    prompt: str


class OpenAIModel(BaseKServeModel):
    """
    An abstract model with methods for implementing OpenAI's endpoints.
    """

    def __init__(self, name: str):
        super().__init__(name)

        # We don't support the `load()` method on OpenAIModel yet
        # Assume the model is ready
        self.ready = True


class OpenAIGenerativeModel(OpenAIModel):
    """
    An abstract model with methods for implementing OpenAI's completions (v1/completions)
    and chat completions (v1/chat/completions) endpoints.

    Users should extend this model and implement the abstract methods in order to expose
    these endpoints.
    """

    @abstractmethod
    async def create_completion(
        self,
        request: CompletionRequest,
        raw_request: Optional[Request] = None,
        context: Optional[Dict[str, Any]] = None,
    ) -> Union[AsyncGenerator[str, None], Completion, ErrorResponse]:
        pass

    @abstractmethod
    async def create_chat_completion(
        self,
        request: ChatCompletionRequest,
        raw_request: Optional[Request] = None,
        context: Optional[Dict[str, Any]] = None,
    ) -> Union[AsyncGenerator[str, None], ChatCompletion, ErrorResponse]:
        pass


class OpenAIEncoderModel(OpenAIModel):
    """
    An abstract model with methods for implementing Embeddings (v1/embeddings) and Rerank (v1/rerank) endpoint.

    Users should extend this model and implement the abstract methods in order to expose
    these endpoints.
    """

    @abstractmethod
    async def create_embedding(
        self,
        request: EmbeddingRequest,
        raw_request: Optional[Request] = None,
        context: Optional[Dict[str, Any]] = None,
    ) -> Union[AsyncGenerator[str, None], Embedding, ErrorResponse]:
        pass

    @abstractmethod
    async def create_rerank(
        self,
        request: RerankRequest,
        raw_request: Optional[Request] = None,
        context: Optional[Dict[str, Any]] = None,
    ) -> Union[AsyncGenerator[str, None], Rerank, ErrorResponse]:
        pass


class AsyncMappingIterator:
    def __init__(
        self,
        iterator: AsyncIterator,
        mapper: Callable = lambda item: item,
        skip_none: bool = True,
        close: Optional[Callable] = None,
    ):
        self.iterator = iterator
        self.mapper = mapper
        self.skip_none = skip_none
        self.close = close

    def __aiter__(self):
        return self

    async def __anext__(self):
        # This will raise StopAsyncIteration when there are no more completions.
        # We don't catch it so it will stop our iterator as well.
        async def next():
            try:
                return self.mapper(await self.iterator.__anext__())
            except Exception as e:
                if self.close:
                    if inspect.iscoroutinefunction(self.close):
                        await self.close()
                    else:
                        self.close()
                raise e

        mapped_item = await next()
        if self.skip_none:
            while mapped_item is None:
                mapped_item = await next()
        return mapped_item



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import tempfile
from unittest import mock

import pytest
from kubernetes.client import V1ObjectMeta, V1ServiceAccount, V1ServiceAccountList, rest

from kserve import constants
from kserve.api.creds_utils import (
    check_sa_exists,
    create_secret,
    create_service_account,
    get_creds_name_from_config_map,
    patch_service_account,
    set_azure_credentials,
    set_gcs_credentials,
    set_s3_credentials,
    set_service_account,
)


@mock.patch("kubernetes.client.CoreV1Api.list_namespaced_service_account")
def test_check_sa_exists(mock_client):
    # Mock kubernetes client to return 2 accounts
    accounts = V1ServiceAccountList(
        items=[V1ServiceAccount(metadata=V1ObjectMeta(name=n)) for n in ["a", "b"]]
    )
    mock_client.return_value = accounts

    # then a, b should exist, c should not
    assert check_sa_exists("kubeflow", "a") is True
    assert check_sa_exists("kubeflow", "b") is True
    assert check_sa_exists("kubeflow", "c") is False


@mock.patch("kubernetes.client.CoreV1Api.create_namespaced_service_account")
def test_create_service_account(mock_client):
    sa_name = "test"
    namespace = "kserve-test"
    secret_name = "test_secret"
    create_service_account(secret_name, namespace, sa_name)
    mock_client.assert_called_once()

    mock_client.side_effect = rest.ApiException("foo")
    with pytest.raises(RuntimeError):
        sa_name = "test"
        namespace = "kserve-test"
        secret_name = "test_secret"
        create_service_account(secret_name, namespace, sa_name)


@mock.patch("kubernetes.client.CoreV1Api.patch_namespaced_service_account")
def test_patch_service_account(mock_client):
    sa_name = "test"
    namespace = "kserve-test"
    secret_name = "test_secret"
    patch_service_account(secret_name, namespace, sa_name)
    mock_client.assert_called_once()

    mock_client.side_effect = rest.ApiException("foo")
    with pytest.raises(RuntimeError):
        sa_name = "test"
        namespace = "kserve-test"
        secret_name = "test_secret"
        patch_service_account(secret_name, namespace, sa_name)


@mock.patch("kubernetes.client.CoreV1Api.create_namespaced_secret")
def test_create_secret(mock_create_secret):
    namespace = "test"
    secret_name = "test-secret"
    mock_create_secret.return_value = mock.Mock(**{"metadata.name": secret_name})
    assert create_secret(namespace) == secret_name

    with pytest.raises(RuntimeError):
        mock_create_secret.side_effect = rest.ApiException("foo")
        create_secret(namespace)


@mock.patch("kserve.api.creds_utils.create_service_account")
@mock.patch("kserve.api.creds_utils.patch_service_account")
@mock.patch("kserve.api.creds_utils.check_sa_exists")
def test_set_service_account(
    mock_check_sa_exists, mock_patch_service_account, mock_create_service_account
):
    namespace = "test"
    service_account = V1ServiceAccount()
    secret_name = "test-secret"
    mock_check_sa_exists.return_value = True
    set_service_account(namespace, service_account, secret_name)
    mock_patch_service_account.assert_called_once()

    mock_check_sa_exists.return_value = False
    set_service_account(namespace, service_account, secret_name)
    mock_create_service_account.assert_called_once()


@mock.patch("kubernetes.client.CoreV1Api.read_namespaced_config_map")
def test_get_creds_name_from_config_map(mock_read_config_map):
    mock_read_config_map.return_value = mock.Mock(
        **{
            "data": {
                "credentials": """{
        "gcs": {"gcsCredentialFileName": "gcs_cred.json"},
        "s3": {"s3AccessKeyIDName": "s3_access_key.json",
               "s3SecretAccessKeyName": "s3_secret.json"}}"""
            }
        }
    )
    test_cases = {
        "gcsCredentialFileName": "gcs_cred.json",
        "s3AccessKeyIDName": "s3_access_key.json",
        "s3SecretAccessKeyName": "s3_secret.json",
    }
    for cred, result in test_cases.items():
        assert get_creds_name_from_config_map(cred) == result

    with pytest.raises(RuntimeError):
        get_creds_name_from_config_map("invalidCred")

    mock_read_config_map.side_effect = rest.ApiException("foo")
    assert get_creds_name_from_config_map("gcsCredentialFileName") is None


@mock.patch("kserve.api.creds_utils.set_service_account")
@mock.patch("kserve.api.creds_utils.create_secret")
@mock.patch("kserve.api.creds_utils.get_creds_name_from_config_map")
def test_set_gcs_credentials(
    mock_get_creds_name, mock_create_secret, mock_set_service_account
):
    namespace = "test"
    service_account = V1ServiceAccount()
    temp_cred_file = tempfile.NamedTemporaryFile(suffix=".json")
    cred_file_name = temp_cred_file.name
    mock_get_creds_name.return_value = cred_file_name
    mock_create_secret.return_value = "test-secret"
    set_gcs_credentials(namespace, cred_file_name, service_account)
    mock_get_creds_name.assert_called()
    mock_create_secret.assert_called()
    mock_set_service_account.assert_called()

    mock_get_creds_name.return_value = None
    set_gcs_credentials(namespace, cred_file_name, service_account)
    mock_get_creds_name.assert_called()
    mock_create_secret.assert_called()
    mock_set_service_account.assert_called()


@mock.patch("kserve.api.creds_utils.set_service_account")
@mock.patch("kserve.api.creds_utils.create_secret")
@mock.patch("kserve.api.creds_utils.get_creds_name_from_config_map")
def test_set_s3_credentials(
    mock_get_creds_name, mock_create_secret, mock_set_service_account
):
    namespace = "test"
    endpoint = "https://s3.aws.com"
    region = "ap-south-1"
    use_https = True
    verfify_ssl = True
    cabundle = "/user/test/cert.pem"
    data = {
        constants.S3_ACCESS_KEY_ID_DEFAULT_NAME: "XXXXXXXXXXXX",
        constants.S3_SECRET_ACCESS_KEY_DEFAULT_NAME: "XXXXXXXXXXXX",
    }
    annotations = {
        constants.KSERVE_GROUP + "/s3-endpoint": endpoint,
        constants.KSERVE_GROUP + "/s3-region": region,
        constants.KSERVE_GROUP + "/s3-usehttps": use_https,
        constants.KSERVE_GROUP + "/s3-verifyssl": verfify_ssl,
        constants.KSERVE_GROUP + "/s3-cabundle": cabundle,
    }
    creds_str = b"""
    [default]
    aws_access_key_id = XXXXXXXXXXXX
    aws_secret_access_key = XXXXXXXXXXXX
    """

    with tempfile.NamedTemporaryFile() as creds_file:
        creds_file.write(creds_str)
        creds_file.seek(0)
        mock_get_creds_name.return_value = None
        mock_create_secret.return_value = "test-secret"
        set_s3_credentials(
            namespace,
            creds_file.name,
            V1ServiceAccount(),
            s3_endpoint=endpoint,
            s3_region=region,
            s3_use_https=use_https,
            s3_verify_ssl=verfify_ssl,
            s3_cabundle=cabundle,
        )
    mock_create_secret.assert_called_with(
        namespace=namespace, annotations=annotations, data=data
    )
    mock_get_creds_name.asset_called()
    mock_set_service_account.assert_called()


@mock.patch("kserve.api.creds_utils.set_service_account")
@mock.patch("kserve.api.creds_utils.create_secret")
def test_set_azure_credentials(mock_create_secret, mock_set_service_account):
    namespace = "test"
    creds = {
        "clientId": "XXXXXXXXXXX",
        "clientSecret": "XXXXXXXXXXX",
        "subscriptionId": "XXXXXXXXXXX",
        "tenantId": "XXXXXXXXXXX",
    }
    data = {
        "AZURE_CLIENT_ID": creds["clientId"],
        "AZURE_CLIENT_SECRET": creds["clientSecret"],
        "AZURE_SUBSCRIPTION_ID": creds["subscriptionId"],
        "AZURE_TENANT_ID": creds["tenantId"],
    }
    with tempfile.NamedTemporaryFile(suffix=".json") as creds_file:
        creds_file.write(json.dumps(creds).encode("utf-8"))
        creds_file.seek(0)
        mock_create_secret.return_value = "test-secret"
        set_azure_credentials(namespace, creds_file.name, V1ServiceAccount())
    mock_create_secret.assert_called_with(namespace=namespace, data=data)
    mock_set_service_account.assert_called()



# Copyright 2025 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import multiprocessing as mp
import os
import signal
from socket import socket
import threading
from typing import Any, Callable, List, Optional, Union

from kserve import logging
from kserve.logging import logger
from kserve.protocol.dataplane import DataPlane
from kserve.protocol.model_repository_extension import ModelRepositoryExtension
from kserve.protocol.rest.server import RESTServer

mp.allow_connection_pickling()
spawn = mp.get_context("spawn")


class RESTServerProcess:
    def __init__(
        self,
        sockets: list[socket],
        target: Callable[[list[socket]], None],
        log_config_file: Optional[str] = None,
    ) -> None:
        self._real_target = target
        # Pipe is used to identify if the process is responsive.
        # A ping request is sent from the parent_conn and is received by the child_conn which in turn responds with a pong.
        # If the request times out, then the process is considered unresponsive. We then kill the process and recreate it.
        self._parent_conn, self._child_conn = mp.Pipe()
        self._process = spawn.Process(target=self.target, args=[sockets])
        self._log_config_file = log_config_file

    def ping(self, timeout: float = 5) -> bool:
        self._parent_conn.send(b"ping")
        if self._parent_conn.poll(timeout):
            self._parent_conn.recv()
            return True
        return False

    def pong(self) -> None:
        self._child_conn.recv()
        self._child_conn.send(b"pong")

    def always_pong(self) -> None:
        while True:
            self.pong()

    def target(self, sockets: list[socket]) -> Any:
        if os.name == "nt":
            # Windows doesn't support SIGTERM, so we use SIGBREAK instead.
            # And then we raise SIGTERM when SIGBREAK is received.
            # https://learn.microsoft.com/zh-cn/cpp/c-runtime-library/reference/signal?view=msvc-170
            signal.signal(
                signal.SIGBREAK,
                lambda sig, frame: signal.raise_signal(signal.SIGTERM),
            )
        # Configure logging for every child
        logging.configure_logging(self._log_config_file)
        threading.Thread(target=self.always_pong, daemon=True).start()
        try:
            return self._real_target(sockets=sockets)
        except KeyboardInterrupt:
            # Suppress side-effects of signal propagation
            # the parent process already expects us to end, so no vital information is lost
            # https://github.com/encode/uvicorn/pull/2317
            pass

    def is_alive(self, timeout: float = 5) -> bool:
        if not self._process.is_alive():
            return False

        return self.ping(timeout)

    def start(self) -> None:
        self._process.start()

    def terminate(self) -> None:
        if self._process.exitcode is None:  # Process is still running
            assert self._process.pid is not None
            if os.name == "nt":
                # Windows doesn't support SIGTERM.
                # So send SIGBREAK, and then in process raise SIGTERM.
                os.kill(self._process.pid, signal.CTRL_BREAK_EVENT)
            else:
                os.kill(self._process.pid, signal.SIGTERM)
        self._parent_conn.close()
        self._child_conn.close()

    def kill(self) -> None:
        # In Windows, the method will call `TerminateProcess` to kill the process.
        # In Unix, the method will send SIGKILL to the process.
        self._process.kill()

    async def wait_for_termination(self, grace_period: Optional[int] = None):
        """Wait for the process to terminate. When a timeout occurs,
        it cancels the task and raises TimeoutError."""

        async def _wait_for_process():
            while self._process.exitcode is None:
                await asyncio.sleep(0.1)

        await asyncio.wait_for(_wait_for_process(), timeout=grace_period)

    @property
    def pid(self) -> Union[int, None]:
        return self._process.pid


class RESTServerMultiProcess:
    def __init__(
        self,
        app: str,
        data_plane: DataPlane,
        model_repository_extension: ModelRepositoryExtension,
        http_port: int,
        access_log_format: Optional[str] = None,
        workers: int = 1,
        grace_period: int = 30,
        log_config_file: Optional[str] = None,
    ) -> None:
        self.log_config_file = log_config_file
        self._rest_server = RESTServer(
            app,
            data_plane,
            model_repository_extension,
            http_port,
            access_log_format,
            workers,
            grace_period,
        )
        self._processes: List[RESTServerProcess] = []
        self.should_exit = asyncio.Event()

    def init_processes(self, sockets: List[socket]) -> None:
        for _ in range(self._rest_server.config.workers):
            p = RESTServerProcess(
                sockets,
                self._rest_server.run,
                log_config_file=self.log_config_file,
            )
            self._processes.append(p)
            p.start()
            logger.info("Started child process [%s]", p.pid)

    async def start(self) -> None:
        logger.info(
            "Starting uvicorn with %s workers", self._rest_server.config.workers
        )
        sockets = [self._rest_server.config.bind_socket()]
        logger.info("Started parent process [%s]", os.getpid())
        self.init_processes(sockets)
        # Blocks until the parent process is ready to exit
        while not self.should_exit.is_set():
            await asyncio.sleep(0.5)
            await self.keep_subprocess_alive(sockets)
        # Propagate signal to all child processes and wait for termination
        await self.terminate_all()

    async def keep_subprocess_alive(self, sockets: List[socket]) -> None:
        if self.should_exit.is_set():
            return  # parent process is exiting, no need to keep subprocess alive

        for idx, process in enumerate(self._processes):
            if not process.is_alive():
                process.kill()  # process is hung, kill it
                await process.wait_for_termination()
                if self.should_exit.is_set():
                    return

                logger.info("Child process [%s] died", process.pid)
                process = RESTServerProcess(
                    sockets,
                    self._rest_server.run,
                )
                process.start()
                self._processes[idx] = process
                logger.info("Started new child process [%s]", process.pid)

    async def terminate_all(self) -> None:
        """Propagate signal to all child processes and wait for termination."""
        for p in self._processes:
            p.terminate()

        async def force_terminate(process) -> None:
            try:
                await process.wait_for_termination(
                    self._rest_server.config.timeout_graceful_shutdown
                )
                logger.info("Terminated child process [%s]", process.pid)
            except asyncio.TimeoutError:
                logger.warning(
                    "Child process [%s] took too long to terminate, force terminating",
                    process.pid,
                )
                process.kill()

        await asyncio.gather(*(force_terminate(p) for p in self._processes))
        logger.info("Stopping parent process [%s]", os.getpid())

    async def stop(self, sig: Optional[int] = None) -> None:
        self.should_exit.set()



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
import os
import pathlib
import re
from unittest import mock
from unittest.mock import patch

import avro
import grpc
import httpx
import pytest
import tomlkit
from cloudevents.conversion import to_binary, to_structured
from cloudevents.http import CloudEvent
from ray import serve
from kserve.protocol.rest.openai.types.openapi import (
    CreateChatCompletionResponse as ChatCompletion,
    CreateChatCompletionStreamResponse as ChatCompletionChunk,
    CreateCompletionResponse as Completion,
)
from typing import AsyncIterator, Union
from kserve.errors import InvalidInput, ModelNotFound
from kserve.model import PredictorProtocol
from kserve.predictor_config import PredictorConfig
from kserve.protocol.dataplane import DataPlane
from kserve import context as kserve_context
from kserve.protocol.rest.openai import CompletionRequest, OpenAIGenerativeModel
from kserve.model_repository import ModelRepository
from kserve.ray import RayModel
from test.test_server import (
    DummyModel,
    dummy_cloud_event,
    DummyCEModel,
    DummyAvroCEModel,
    DummyServeModel,
)


@pytest.mark.asyncio
class TestDataPlane:
    MODEL_NAME = "TestModel"

    # Adding two 'params' to the below fixture causes all the tests using this fixture to run twice.
    # First time this fixture uses 'Model' class.
    # The second time it uses the 'RayServeHandle' class.
    @pytest.fixture(scope="class", params=["TEST_RAW_MODEL", "TEST_RAY_SERVE_MODEL"])
    def dataplane_with_model(self, request):  # pylint: disable=no-self-use
        dataplane = DataPlane(model_registry=ModelRepository())

        if request.param == "TEST_RAW_MODEL":
            model = DummyModel(self.MODEL_NAME)
            model.load()
            dataplane._model_registry.update(model)
            yield dataplane
        else:  # request.param == "TEST_RAY_SERVE_MODEL"
            serve.start(http_options={"host": "0.0.0.0", "port": 9071})

            # https://github.com/ray-project/ray/blob/releases/2.8.0/python/ray/serve/deployment.py#L256
            app = DummyServeModel.bind(name=self.MODEL_NAME)
            handle = serve.run(app, name=self.MODEL_NAME, route_prefix="/")

            model = RayModel(self.MODEL_NAME, handle=handle)
            model.load()
            dataplane._model_registry.update(model)
            yield dataplane
            serve.delete(name="TestModel")

    async def test_get_model_from_registry(self):
        dataplane = DataPlane(model_registry=ModelRepository())
        model_name = "FakeModel"
        with pytest.raises(ModelNotFound) as http_exec:
            dataplane.get_model_from_registry(model_name)
        # assert http_exec.value.status_code == 404
        assert http_exec.value.reason == f"Model with name {model_name} does not exist."

        ready_model = DummyModel("Model")
        ready_model.load()
        dataplane._model_registry.update(ready_model)
        dataplane.get_model_from_registry("Model")

    async def test_liveness(self):
        assert (await DataPlane.live()) == {"status": "alive"}

    async def test_server_readiness(self, dataplane_with_model):
        assert (await dataplane_with_model.ready()) is True

        not_ready_model = DummyModel("NotReadyModel")
        # model.load()  # Model not loaded, i.e. model not ready
        dataplane_with_model._model_registry.update(not_ready_model)
        # The model server readiness endpoint should return 'True' irrespective of the readiness
        # of the model loaded into it.
        assert (await dataplane_with_model.ready()) is True

    async def test_model_readiness(self):
        dataplane = DataPlane(model_registry=ModelRepository())

        ready_model = DummyModel("ReadyModel")
        ready_model.load()
        dataplane._model_registry.update(ready_model)
        is_ready = await dataplane.model_ready(ready_model.name)
        assert is_ready is True

        not_ready_model = DummyModel("NotReadyModel")
        # model.load()  # Model not loaded, i.e. not ready
        dataplane._model_registry.update(not_ready_model)
        is_ready = await dataplane.model_ready(not_ready_model.name)
        assert is_ready is False

    async def test_server_metadata(self):
        with open(pathlib.Path(__file__).parent.parent / "pyproject.toml") as toml_file:
            toml_config = tomlkit.load(toml_file)
            version = toml_config["project"]["version"].strip()

        dataplane = DataPlane(model_registry=ModelRepository())
        expected_metadata = {
            "name": "kserve",
            "version": version,
            "extensions": ["model_repository_extension"],
        }
        assert dataplane.metadata() == expected_metadata

    async def test_model_metadata(self, dataplane_with_model):
        assert await dataplane_with_model.model_metadata(self.MODEL_NAME) == {
            "name": self.MODEL_NAME,
            "platform": "",
            "inputs": [],
            "outputs": [],
        }

    async def test_infer(self, dataplane_with_model):
        body = b'{"instances":[[1,2]]}'
        infer_request, req_attributes = dataplane_with_model.decode(body, None)
        resp, headers = await dataplane_with_model.infer(self.MODEL_NAME, infer_request)
        resp, headers = dataplane_with_model.encode(
            self.MODEL_NAME, resp, headers, req_attributes
        )
        assert (resp, headers) == ({"predictions": [[1, 2]]}, {})  # body

    async def test_explain(self, dataplane_with_model: DataPlane):
        body = b'{"instances":[[1,2]]}'
        infer_request, req_attributes = dataplane_with_model.decode(body, None)
        resp, headers = await dataplane_with_model.explain(
            self.MODEL_NAME, infer_request
        )
        resp, headers = dataplane_with_model.encode(
            self.MODEL_NAME, resp, headers, req_attributes
        )
        assert (resp, headers) == ({"predictions": [[1, 2]]}, {})


@pytest.mark.asyncio
class TestDataPlaneCloudEvent:
    MODEL_NAME = "TestModel"

    @pytest.fixture(scope="class")
    def dataplane_with_ce_model(self):  # pylint: disable=no-self-use
        dataplane = DataPlane(model_registry=ModelRepository())
        model = DummyCEModel(self.MODEL_NAME)
        model.load()
        dataplane._model_registry.update(model)
        return dataplane

    async def test_infer_ce_structured(self, dataplane_with_ce_model: DataPlane):
        event: CloudEvent = dummy_cloud_event({"instances": [[1, 2]]})
        headers, body = to_structured(event)
        infer_request, req_attributes = dataplane_with_ce_model.decode(body, headers)
        resp, response_headers = await dataplane_with_ce_model.infer(
            self.MODEL_NAME, infer_request, headers
        )
        resp, res_headers = dataplane_with_ce_model.encode(
            self.MODEL_NAME, resp, headers, req_attributes
        )
        response_headers.update(res_headers)
        body = json.loads(resp)

        assert response_headers["content-type"] == "application/cloudevents+json"

        assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert body["data"] == {"predictions": [[1, 2]]}
        assert body["specversion"] == "1.0"
        assert body["source"] == "io.kserve.inference.TestModel"
        assert body["type"] == "io.kserve.inference.response"
        assert body["time"] > "2021-01-28T21:04:43.144141+00:00"

    async def test_infer_custom_ce_attributes(self, dataplane_with_ce_model):
        with mock.patch.dict(
            os.environ,
            {
                "CE_SOURCE": "io.kserve.inference.CustomSource",
                "CE_TYPE": "io.kserve.custom_type",
            },
        ):
            event = dummy_cloud_event({"instances": [[1, 2]]})
            headers, body = to_structured(event)

            infer_request, req_attributes = dataplane_with_ce_model.decode(
                body, headers
            )
            resp, response_headers = await dataplane_with_ce_model.infer(
                self.MODEL_NAME, infer_request, headers
            )
            resp, res_headers = dataplane_with_ce_model.encode(
                self.MODEL_NAME, resp, headers, req_attributes
            )
            response_headers.update(res_headers)
            body = json.loads(resp)

            assert response_headers["content-type"] == "application/cloudevents+json"

            assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            assert body["data"] == {"predictions": [[1, 2]]}
            assert body["source"] == "io.kserve.inference.CustomSource"
            assert body["type"] == "io.kserve.custom_type"

    async def test_infer_merge_structured_ce_attributes(
        self, dataplane_with_ce_model: DataPlane
    ):
        with mock.patch.dict(os.environ, {"CE_MERGE": "true"}):
            event = dummy_cloud_event({"instances": [[1, 2]]}, add_extension=True)
            headers, body = to_structured(event)

            infer_request, req_attributes = dataplane_with_ce_model.decode(
                body, headers
            )
            resp, response_headers = await dataplane_with_ce_model.infer(
                self.MODEL_NAME, infer_request, headers
            )
            resp, res_headers = dataplane_with_ce_model.encode(
                self.MODEL_NAME, resp, headers, req_attributes
            )
            response_headers.update(res_headers)
            body = json.loads(resp)

            assert response_headers["content-type"] == "application/cloudevents+json"

            assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            assert body["data"] == {"predictions": [[1, 2]]}
            assert body["source"] == "io.kserve.inference.TestModel"
            assert body["type"] == "io.kserve.inference.response"
            assert (
                body["custom-extension"] == "custom-value"
            )  # Added by add_extension=True in dummy_cloud_event
            assert body["time"] > "2021-01-28T21:04:43.144141+00:00"

    async def test_infer_merge_binary_ce_attributes(self, dataplane_with_ce_model):
        with mock.patch.dict(os.environ, {"CE_MERGE": "true"}):
            event = dummy_cloud_event(
                {"instances": [[1, 2]]}, set_contenttype=True, add_extension=True
            )
            headers, body = to_binary(event)

            infer_request, req_attributes = dataplane_with_ce_model.decode(
                body, headers
            )
            resp, response_headers = await dataplane_with_ce_model.infer(
                self.MODEL_NAME, infer_request, headers
            )
            resp, res_headers = dataplane_with_ce_model.encode(
                self.MODEL_NAME, resp, headers, req_attributes
            )
            response_headers.update(res_headers)

            assert response_headers["content-type"] == "application/json"
            assert response_headers["ce-specversion"] == "1.0"
            assert response_headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            # Added by add_extension=True in dummy_cloud_event
            assert response_headers["ce-custom-extension"] == "custom-value"
            assert response_headers["ce-source"] == "io.kserve.inference.TestModel"
            assert response_headers["ce-type"] == "io.kserve.inference.response"
            assert response_headers["ce-time"] > "2021-01-28T21:04:43.144141+00:00"
            assert resp == b'{"predictions": [[1, 2]]}'

    async def test_infer_ce_binary_dict(self, dataplane_with_ce_model):
        event = dummy_cloud_event({"instances": [[1, 2]]}, set_contenttype=True)
        headers, body = to_binary(event)

        infer_request, req_attributes = dataplane_with_ce_model.decode(body, headers)
        resp, response_headers = await dataplane_with_ce_model.infer(
            self.MODEL_NAME, infer_request, headers
        )
        resp, res_headers = dataplane_with_ce_model.encode(
            self.MODEL_NAME, resp, headers, req_attributes
        )
        response_headers.update(res_headers)

        assert response_headers["content-type"] == "application/json"
        assert response_headers["ce-specversion"] == "1.0"
        assert response_headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert response_headers["ce-source"] == "io.kserve.inference.TestModel"
        assert response_headers["ce-type"] == "io.kserve.inference.response"
        assert response_headers["ce-time"] > "2021-01-28T21:04:43.144141+00:00"
        assert resp == b'{"predictions": [[1, 2]]}'

    async def test_infer_ce_binary_bytes(self, dataplane_with_ce_model):
        event = dummy_cloud_event(b'{"instances":[[1,2]]}', set_contenttype=True)
        headers, body = to_binary(event)

        infer_request, req_attributes = dataplane_with_ce_model.decode(body, headers)
        resp, response_headers = await dataplane_with_ce_model.infer(
            self.MODEL_NAME, infer_request, headers
        )
        resp, res_headers = dataplane_with_ce_model.encode(
            self.MODEL_NAME, resp, headers, req_attributes
        )
        response_headers.update(res_headers)
        assert response_headers["content-type"] == "application/json"
        assert response_headers["ce-specversion"] == "1.0"
        assert response_headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert response_headers["ce-source"] == "io.kserve.inference.TestModel"
        assert response_headers["ce-type"] == "io.kserve.inference.response"
        assert response_headers["ce-time"] > "2021-01-28T21:04:43.144141+00:00"
        assert resp == b'{"predictions": [[1, 2]]}'

    async def test_infer_ce_bytes_bad_format_exception(self, dataplane_with_ce_model):
        event = dummy_cloud_event(b"{", set_contenttype=True)
        headers, body = to_binary(event)

        with pytest.raises(InvalidInput) as err:
            infer_request, req_attributes = dataplane_with_ce_model.decode(
                body, headers
            )
            await dataplane_with_ce_model.infer(self.MODEL_NAME, infer_request, headers)

        error_regex = re.compile(
            "Failed to decode or parse binary json cloudevent: "
            "unexpected end of data:*"
        )
        assert error_regex.match(err.value.reason) is not None

    async def test_infer_ce_bytes_bad_hex_format_exception(
        self, dataplane_with_ce_model
    ):
        event = dummy_cloud_event(b"0\x80\x80\x06World!\x00\x00", set_contenttype=True)
        headers, body = to_binary(event)

        with pytest.raises(InvalidInput) as err:
            infer_request, req_attributes = dataplane_with_ce_model.decode(
                body, headers
            )
            await dataplane_with_ce_model.infer(self.MODEL_NAME, infer_request, headers)

        error_regex = re.compile(
            "Failed to decode or parse binary json cloudevent: 'utf-8' codec "
            "can't decode byte 0x80 in position 1: invalid start byte.*"
        )

        assert error_regex.match(err.value.reason) is not None


@pytest.mark.asyncio
class TestDataPlaneAvroCloudEvent:
    test_avsc_schema = """
            {
            "namespace": "example.avro",
             "type": "record",
             "name": "User",
             "fields": [
                 {"name": "name", "type": "string"},
                 {"name": "favorite_number",  "type": ["int", "null"]},
                 {"name": "favorite_color", "type": ["string", "null"]}
             ]
            }
            """

    MODEL_NAME = "TestModel"

    @pytest.fixture(scope="class")
    def dataplane_with_ce_model(self):  # pylint: disable=no-self-use
        dataplane = DataPlane(model_registry=ModelRepository())
        model = DummyAvroCEModel(self.MODEL_NAME)
        model.load()
        dataplane._model_registry.update(model)
        return dataplane

    async def test_infer_ce_avro_binary(self, dataplane_with_ce_model):
        schema = avro.schema.parse(self.test_avsc_schema)
        msg = {"name": "foo", "favorite_number": 1, "favorite_color": "pink"}

        writer = avro.io.DatumWriter(schema)
        bytes_writer = io.BytesIO()
        encoder = avro.io.BinaryEncoder(bytes_writer)
        writer.write(msg, encoder)
        data = bytes_writer.getvalue()

        event = dummy_cloud_event(
            data, set_contenttype=True, contenttype="application/avro"
        )
        # Creates the HTTP request representation of the CloudEvent in binary content mode
        headers, body = to_binary(event)

        infer_request, req_attributes = dataplane_with_ce_model.decode(body, headers)
        resp, response_headers = await dataplane_with_ce_model.infer(
            self.MODEL_NAME, infer_request, headers
        )

        resp, res_headers = dataplane_with_ce_model.encode(
            self.MODEL_NAME, resp, headers, req_attributes
        )
        response_headers.update(res_headers)

        assert response_headers["content-type"] == "application/json"
        assert response_headers["ce-specversion"] == "1.0"
        assert response_headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert response_headers["ce-source"] == "io.kserve.inference.TestModel"
        assert response_headers["ce-type"] == "io.kserve.inference.response"
        assert response_headers["ce-time"] > "2021-01-28T21:04:43.144141+00:00"
        assert resp == b'{"predictions": [["foo", 1, "pink"]]}'


@pytest.mark.asyncio
class TestDataPlaneOpenAI:
    MODEL_NAME = "TestModel"

    class DummyOpenAIModel(OpenAIGenerativeModel):
        async def create_completion(
            self, params: CompletionRequest
        ) -> Union[Completion, AsyncIterator[Completion]]:
            pass

        async def create_chat_completion(
            self, params: CompletionRequest
        ) -> Union[ChatCompletion, AsyncIterator[ChatCompletionChunk]]:
            pass

    async def test_infer_on_openai_completion_model_raises(self):
        openai_model = self.DummyOpenAIModel(self.MODEL_NAME)
        repo = ModelRepository()
        repo.update(openai_model)
        dataplane = DataPlane(model_registry=repo)

        with pytest.raises(ValueError) as exc:
            await dataplane.infer(
                model_name=self.MODEL_NAME,
                request={},
            )

        assert (
            exc.value.args[0]
            == "Model of type DummyOpenAIModel does not support inference"
        )


@pytest.mark.asyncio
class TestDataplaneTransformer:

    async def test_dataplane_rest_with_ssl_enabled(self, httpx_mock):
        # scenario: getting a 2xx response from predictor with ssl enabled
        predictor_host = "ready.host"
        httpx_mock.add_response(
            url=re.compile(f"https://{predictor_host}/*"), json={"status": "alive"}
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V1.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
            predictor_use_ssl=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())

        assert (await dataplane.ready()) is True

    @patch("kserve.protocol.dataplane.InferenceClientFactory.get_grpc_client")
    async def test_dataplane_grpc_with_ssl_enabled(self, mock_grpc_client):
        # scenario: getting a 2xx response from predictor with ssl enabled
        predictor_host = "ready.host"
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.GRPC_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
            predictor_use_ssl=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        mock_is_server_ready = mock.AsyncMock(return_value=True)
        mock_grpc_client.return_value.is_server_ready = mock_is_server_ready
        assert (await dataplane.ready()) is True
        mock_grpc_client.assert_called_with(
            url=predictor_host, timeout=5, retries=2, use_ssl=True
        )

    async def test_server_readiness_v1(self, httpx_mock):
        # scenario: getting a 2xx response from predictor
        predictor_host = "ready.host"
        httpx_mock.add_response(
            url=re.compile(f"http://{predictor_host}/*"), json={"status": "alive"}
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V1.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        assert (await dataplane.ready()) is True

        # scenario: not a 2xx response from predictor
        predictor_host = "not-ready.host"
        httpx_mock.add_response(
            url=re.compile(f"http://{predictor_host}/*"), status_code=500
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V1.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        with pytest.raises(httpx.HTTPStatusError):
            await dataplane.ready()

    async def test_server_readiness_v2(self, httpx_mock):
        # scenario: getting a 2xx response from predictor
        predictor_host = "ready.host"
        httpx_mock.add_response(
            url=re.compile(f"http://{predictor_host}/v2/*"), json={"ready": True}
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        assert (await dataplane.ready()) is True

        # scenario: getting a 2xx response from predictor and server not ready
        predictor_host = "not-ready.host"
        httpx_mock.add_response(
            url=re.compile(f"http://{predictor_host}/v2/*"), json={"ready": False}
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        assert (await dataplane.ready()) is False

        # scenario: not a 2xx response from predictor
        predictor_host = "not-ready.host"
        httpx_mock.add_response(
            url=re.compile(f"http://{predictor_host}/v2/*"), status_code=500
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        with pytest.raises(httpx.HTTPStatusError):
            await dataplane.ready()

    @patch("kserve.protocol.dataplane.InferenceClientFactory.get_grpc_client")
    async def test_server_readiness_grpc_v2(self, mock_grpc_client):
        # scenario: getting a 2xx response from predictor
        predictor_host = "ready.host"
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.GRPC_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        mock_is_server_ready = mock.AsyncMock(return_value=True)
        mock_grpc_client.return_value.is_server_ready = mock_is_server_ready
        assert (await dataplane.ready()) is True
        mock_grpc_client.assert_called_with(
            url=predictor_host, timeout=5, retries=2, use_ssl=False
        )

        # scenario: getting a 2xx response from predictor and server not ready
        predictor_host = "not-ready.host"
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.GRPC_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        mock_is_server_ready = mock.AsyncMock(return_value=False)
        mock_grpc_client.return_value.is_server_ready = mock_is_server_ready
        assert (await dataplane.ready()) is False
        mock_grpc_client.assert_called_with(
            url=predictor_host, timeout=5, retries=2, use_ssl=False
        )

    async def test_model_readiness_v1(self, httpx_mock):
        # scenario: getting a 2xx response from predictor
        predictor_host = "ready.host"
        httpx_mock.add_response(
            url=re.compile(f"http://{predictor_host}/v1/*"), json={"ready": True}
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V1.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        # Transformer model ready
        ready_model = DummyModel("ReadyModel")
        dataplane._model_registry.update(ready_model)
        assert (await dataplane.model_ready(ready_model.name)) is True

        # scenario: getting a 2xx response from predictor and model not ready
        predictor_host = "ready.host"
        httpx_mock.add_response(
            url=re.compile(f"http://{predictor_host}/v1/*"), json={"ready": False}
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V1.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        # Transformer model ready
        not_ready_model = DummyModel("NotReadyModel")
        dataplane._model_registry.update(not_ready_model)
        assert (await dataplane.model_ready(not_ready_model.name)) is False

        # scenario: not a 2xx response from predictor
        predictor_host = "not-ready.host"
        httpx_mock.add_response(
            url=re.compile(f"http://{predictor_host}/v1/*"), status_code=503
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V1.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        # Transformer model is not ready
        not_ready_model = DummyModel("NotReadyModel")
        dataplane._model_registry.update(not_ready_model)
        assert await dataplane.model_ready(not_ready_model.name) is False

    async def test_model_readiness_v2(self, httpx_mock):
        # scenario: getting a 2xx response from predictor
        predictor_host = "ready.host"
        httpx_mock.add_response(url=re.compile(f"http://{predictor_host}/v2/*"))
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        # Transformer model ready
        ready_model = DummyModel("ReadyModel")
        dataplane._model_registry.update(ready_model)
        assert (await dataplane.model_ready(ready_model.name)) is True

        # scenario: getting a 400 response from predictor and model not ready
        predictor_host = "triton-not-ready.host"
        # Triton returns a non-200 response if model is not ready
        httpx_mock.add_response(
            url=re.compile(f"http://{predictor_host}/v2/*"), status_code=400
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        # Transformer model ready
        not_ready_model = DummyModel("NotReadyModel")
        dataplane._model_registry.update(not_ready_model)
        assert (await dataplane.model_ready(not_ready_model.name)) is False

        # scenario: 503 response from predictor
        # According to V2 protocol, 200 status code indicates true and a 4xx status code indicates false.
        # The HTTP response body should be empty.
        # However, KServe returns 503 when not ready
        predictor_host = "not-ready.host"
        httpx_mock.add_response(
            url=re.compile(f"http://{predictor_host}/v2/*"), status_code=503
        )
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        # Transformer model is not ready
        not_ready_model = DummyModel("NotReadyModel")
        dataplane._model_registry.update(not_ready_model)
        assert await dataplane.model_ready(not_ready_model.name) is False

        # Connection error
        predictor_host = "not-reachable.host"
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        # Transformer model is not ready
        not_ready_model = DummyModel("NotReadyModel")
        dataplane._model_registry.update(not_ready_model)
        httpx_mock.add_exception(
            url=re.compile(f"http://{predictor_host}/v2/*"),
            exception=httpx.ConnectError("All connection attempts failed"),
        )
        assert (await dataplane.model_ready(not_ready_model.name)) is False

    @patch("kserve.protocol.dataplane.InferenceClientFactory.get_grpc_client")
    async def test_model_readiness_grpc_v2(self, mock_grpc_client):
        # scenario: getting a 2xx response from predictor
        predictor_host = "ready.host"
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.GRPC_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        # Transformer model ready
        ready_model = DummyModel("ReadyModel")
        dataplane._model_registry.update(ready_model)
        mock_is_model_ready = mock.AsyncMock(return_value=True)
        mock_grpc_client.return_value.is_model_ready = mock_is_model_ready
        assert (await dataplane.model_ready(ready_model.name)) is True
        mock_grpc_client.assert_called_with(
            url=predictor_host, timeout=5, retries=2, use_ssl=False
        )

        # scenario: getting a 2xx response from predictor and server not ready
        predictor_host = "not-ready.host"
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.GRPC_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        # Transformer model is not ready
        not_ready_model = DummyModel("NotReadyModel")
        dataplane._model_registry.update(not_ready_model)
        mock_is_model_ready = mock.AsyncMock(return_value=False)
        mock_grpc_client.return_value.is_model_ready = mock_is_model_ready
        assert (await dataplane.model_ready(not_ready_model.name)) is False
        mock_grpc_client.assert_called_with(
            url=predictor_host, timeout=5, retries=2, use_ssl=False
        )

        # Connection error
        predictor_host = "not-reachable.host"
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.GRPC_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=True,
        )
        kserve_context.set_predictor_config(predictor_config)
        dataplane = DataPlane(model_registry=ModelRepository())
        dataplane._model_registry.update(ready_model)
        mock_grpc_client.side_effect = grpc.RpcError("Mocked exception")
        assert (await dataplane.model_ready(ready_model.name)) is False

    @patch("kserve.protocol.dataplane.InferenceClientFactory.get_grpc_client")
    @patch("kserve.protocol.dataplane.InferenceClientFactory.get_rest_client")
    async def test_dataplane_with_predictor_health_check_false(
        self, mock_rest_client, mock_grpc_client
    ):
        # Inference client should not be created when predictor_health_check is False
        predictor_host = "ready.host"
        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.GRPC_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=False,
        )
        kserve_context.set_predictor_config(predictor_config)
        _ = DataPlane(model_registry=ModelRepository())
        mock_grpc_client.assert_not_called()

        predictor_config = PredictorConfig(
            predictor_host=predictor_host,
            predictor_protocol=PredictorProtocol.REST_V2.value,
            predictor_request_retries=2,
            predictor_request_timeout_seconds=5,
            predictor_health_check=False,
        )
        kserve_context.set_predictor_config(predictor_config)
        _ = DataPlane(model_registry=ModelRepository())
        mock_rest_client.assert_not_called()



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kserve.configuration import Configuration


class V1beta1ScalerSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'max_replicas': 'int',
        'metric_backend': 'str',
        'metric_query': 'str',
        'min_replicas': 'int',
        'query_parameters': 'str',
        'query_time': 'str',
        'scale_metric': 'str',
        'scale_metric_type': 'str',
        'scale_target': 'int',
        'server_address': 'str'
    }

    attribute_map = {
        'max_replicas': 'maxReplicas',
        'metric_backend': 'metricBackend',
        'metric_query': 'metricQuery',
        'min_replicas': 'minReplicas',
        'query_parameters': 'queryParameters',
        'query_time': 'queryTime',
        'scale_metric': 'scaleMetric',
        'scale_metric_type': 'scaleMetricType',
        'scale_target': 'scaleTarget',
        'server_address': 'serverAddress'
    }

    def __init__(self, max_replicas=None, metric_backend=None, metric_query=None, min_replicas=None, query_parameters=None, query_time=None, scale_metric=None, scale_metric_type=None, scale_target=None, server_address=None, local_vars_configuration=None):  # noqa: E501
        """V1beta1ScalerSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._max_replicas = None
        self._metric_backend = None
        self._metric_query = None
        self._min_replicas = None
        self._query_parameters = None
        self._query_time = None
        self._scale_metric = None
        self._scale_metric_type = None
        self._scale_target = None
        self._server_address = None
        self.discriminator = None

        if max_replicas is not None:
            self.max_replicas = max_replicas
        if metric_backend is not None:
            self.metric_backend = metric_backend
        if metric_query is not None:
            self.metric_query = metric_query
        if min_replicas is not None:
            self.min_replicas = min_replicas
        if query_parameters is not None:
            self.query_parameters = query_parameters
        if query_time is not None:
            self.query_time = query_time
        if scale_metric is not None:
            self.scale_metric = scale_metric
        if scale_metric_type is not None:
            self.scale_metric_type = scale_metric_type
        if scale_target is not None:
            self.scale_target = scale_target
        if server_address is not None:
            self.server_address = server_address

    @property
    def max_replicas(self):
        """Gets the max_replicas of this V1beta1ScalerSpec.  # noqa: E501

        Maximum number of replicas for autoscaling.  # noqa: E501

        :return: The max_replicas of this V1beta1ScalerSpec.  # noqa: E501
        :rtype: int
        """
        return self._max_replicas

    @max_replicas.setter
    def max_replicas(self, max_replicas):
        """Sets the max_replicas of this V1beta1ScalerSpec.

        Maximum number of replicas for autoscaling.  # noqa: E501

        :param max_replicas: The max_replicas of this V1beta1ScalerSpec.  # noqa: E501
        :type: int
        """

        self._max_replicas = max_replicas

    @property
    def metric_backend(self):
        """Gets the metric_backend of this V1beta1ScalerSpec.  # noqa: E501

        MetricsBackend defines the scaling metric type watched by autoscaler possible values are prometheus, graphite.  # noqa: E501

        :return: The metric_backend of this V1beta1ScalerSpec.  # noqa: E501
        :rtype: str
        """
        return self._metric_backend

    @metric_backend.setter
    def metric_backend(self, metric_backend):
        """Sets the metric_backend of this V1beta1ScalerSpec.

        MetricsBackend defines the scaling metric type watched by autoscaler possible values are prometheus, graphite.  # noqa: E501

        :param metric_backend: The metric_backend of this V1beta1ScalerSpec.  # noqa: E501
        :type: str
        """

        self._metric_backend = metric_backend

    @property
    def metric_query(self):
        """Gets the metric_query of this V1beta1ScalerSpec.  # noqa: E501

        Query to run to get metrics from MetricsBackend  # noqa: E501

        :return: The metric_query of this V1beta1ScalerSpec.  # noqa: E501
        :rtype: str
        """
        return self._metric_query

    @metric_query.setter
    def metric_query(self, metric_query):
        """Sets the metric_query of this V1beta1ScalerSpec.

        Query to run to get metrics from MetricsBackend  # noqa: E501

        :param metric_query: The metric_query of this V1beta1ScalerSpec.  # noqa: E501
        :type: str
        """

        self._metric_query = metric_query

    @property
    def min_replicas(self):
        """Gets the min_replicas of this V1beta1ScalerSpec.  # noqa: E501

        Minimum number of replicas, defaults to 1 but can be set to 0 to enable scale-to-zero.  # noqa: E501

        :return: The min_replicas of this V1beta1ScalerSpec.  # noqa: E501
        :rtype: int
        """
        return self._min_replicas

    @min_replicas.setter
    def min_replicas(self, min_replicas):
        """Sets the min_replicas of this V1beta1ScalerSpec.

        Minimum number of replicas, defaults to 1 but can be set to 0 to enable scale-to-zero.  # noqa: E501

        :param min_replicas: The min_replicas of this V1beta1ScalerSpec.  # noqa: E501
        :type: int
        """

        self._min_replicas = min_replicas

    @property
    def query_parameters(self):
        """Gets the query_parameters of this V1beta1ScalerSpec.  # noqa: E501

        A comma-separated list of query Parameters to include while querying the MetricsBackend endpoint.  # noqa: E501

        :return: The query_parameters of this V1beta1ScalerSpec.  # noqa: E501
        :rtype: str
        """
        return self._query_parameters

    @query_parameters.setter
    def query_parameters(self, query_parameters):
        """Sets the query_parameters of this V1beta1ScalerSpec.

        A comma-separated list of query Parameters to include while querying the MetricsBackend endpoint.  # noqa: E501

        :param query_parameters: The query_parameters of this V1beta1ScalerSpec.  # noqa: E501
        :type: str
        """

        self._query_parameters = query_parameters

    @property
    def query_time(self):
        """Gets the query_time of this V1beta1ScalerSpec.  # noqa: E501

        queryTime is relative time range to execute query against. specialized for graphite (https://graphite-api.readthedocs.io/en/latest/api.html#from-until)  # noqa: E501

        :return: The query_time of this V1beta1ScalerSpec.  # noqa: E501
        :rtype: str
        """
        return self._query_time

    @query_time.setter
    def query_time(self, query_time):
        """Sets the query_time of this V1beta1ScalerSpec.

        queryTime is relative time range to execute query against. specialized for graphite (https://graphite-api.readthedocs.io/en/latest/api.html#from-until)  # noqa: E501

        :param query_time: The query_time of this V1beta1ScalerSpec.  # noqa: E501
        :type: str
        """

        self._query_time = query_time

    @property
    def scale_metric(self):
        """Gets the scale_metric of this V1beta1ScalerSpec.  # noqa: E501

        ScaleMetric defines the scaling metric type watched by autoscaler possible values are concurrency, rps, cpu, memory. concurrency, rps are supported via Knative Pod Autoscaler(https://knative.dev/docs/serving/autoscaling/autoscaling-metrics).  # noqa: E501

        :return: The scale_metric of this V1beta1ScalerSpec.  # noqa: E501
        :rtype: str
        """
        return self._scale_metric

    @scale_metric.setter
    def scale_metric(self, scale_metric):
        """Sets the scale_metric of this V1beta1ScalerSpec.

        ScaleMetric defines the scaling metric type watched by autoscaler possible values are concurrency, rps, cpu, memory. concurrency, rps are supported via Knative Pod Autoscaler(https://knative.dev/docs/serving/autoscaling/autoscaling-metrics).  # noqa: E501

        :param scale_metric: The scale_metric of this V1beta1ScalerSpec.  # noqa: E501
        :type: str
        """

        self._scale_metric = scale_metric

    @property
    def scale_metric_type(self):
        """Gets the scale_metric_type of this V1beta1ScalerSpec.  # noqa: E501

        Type of metric to use. Options are Utilization, or AverageValue.  # noqa: E501

        :return: The scale_metric_type of this V1beta1ScalerSpec.  # noqa: E501
        :rtype: str
        """
        return self._scale_metric_type

    @scale_metric_type.setter
    def scale_metric_type(self, scale_metric_type):
        """Sets the scale_metric_type of this V1beta1ScalerSpec.

        Type of metric to use. Options are Utilization, or AverageValue.  # noqa: E501

        :param scale_metric_type: The scale_metric_type of this V1beta1ScalerSpec.  # noqa: E501
        :type: str
        """

        self._scale_metric_type = scale_metric_type

    @property
    def scale_target(self):
        """Gets the scale_target of this V1beta1ScalerSpec.  # noqa: E501

        ScaleTarget specifies the integer target value of the metric type the Autoscaler watches for. concurrency and rps targets are supported by Knative Pod Autoscaler (https://knative.dev/docs/serving/autoscaling/autoscaling-targets/).  # noqa: E501

        :return: The scale_target of this V1beta1ScalerSpec.  # noqa: E501
        :rtype: int
        """
        return self._scale_target

    @scale_target.setter
    def scale_target(self, scale_target):
        """Sets the scale_target of this V1beta1ScalerSpec.

        ScaleTarget specifies the integer target value of the metric type the Autoscaler watches for. concurrency and rps targets are supported by Knative Pod Autoscaler (https://knative.dev/docs/serving/autoscaling/autoscaling-targets/).  # noqa: E501

        :param scale_target: The scale_target of this V1beta1ScalerSpec.  # noqa: E501
        :type: int
        """

        self._scale_target = scale_target

    @property
    def server_address(self):
        """Gets the server_address of this V1beta1ScalerSpec.  # noqa: E501

        Address of MetricsBackend server.  # noqa: E501

        :return: The server_address of this V1beta1ScalerSpec.  # noqa: E501
        :rtype: str
        """
        return self._server_address

    @server_address.setter
    def server_address(self, server_address):
        """Sets the server_address of this V1beta1ScalerSpec.

        Address of MetricsBackend server.  # noqa: E501

        :param server_address: The server_address of this V1beta1ScalerSpec.  # noqa: E501
        :type: str
        """

        self._server_address = server_address

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1beta1ScalerSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1beta1ScalerSpec):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: grpc_predict_v2.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x15grpc_predict_v2.proto\x12\tinference"\x13\n\x11ServerLiveRequest""\n\x12ServerLiveResponse\x12\x0c\n\x04live\x18\x01 \x01(\x08"\x14\n\x12ServerReadyRequest"$\n\x13ServerReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08"2\n\x11ModelReadyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t"#\n\x12ModelReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08"\x17\n\x15ServerMetadataRequest"K\n\x16ServerMetadataResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x12\n\nextensions\x18\x03 \x03(\t"5\n\x14ModelMetadataRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t"\x8d\x02\n\x15ModelMetadataResponse\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08versions\x18\x02 \x03(\t\x12\x10\n\x08platform\x18\x03 \x01(\t\x12?\n\x06inputs\x18\x04 \x03(\x0b\x32/.inference.ModelMetadataResponse.TensorMetadata\x12@\n\x07outputs\x18\x05 \x03(\x0b\x32/.inference.ModelMetadataResponse.TensorMetadata\x1a?\n\x0eTensorMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03"\xee\x06\n\x11ModelInferRequest\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x15\n\rmodel_version\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\x12@\n\nparameters\x18\x04 \x03(\x0b\x32,.inference.ModelInferRequest.ParametersEntry\x12=\n\x06inputs\x18\x05 \x03(\x0b\x32-.inference.ModelInferRequest.InferInputTensor\x12H\n\x07outputs\x18\x06 \x03(\x0b\x32\x37.inference.ModelInferRequest.InferRequestedOutputTensor\x12\x1a\n\x12raw_input_contents\x18\x07 \x03(\x0c\x1a\x94\x02\n\x10InferInputTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03\x12Q\n\nparameters\x18\x04 \x03(\x0b\x32=.inference.ModelInferRequest.InferInputTensor.ParametersEntry\x12\x30\n\x08\x63ontents\x18\x05 \x01(\x0b\x32\x1e.inference.InferTensorContents\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01\x1a\xd5\x01\n\x1aInferRequestedOutputTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12[\n\nparameters\x18\x02 \x03(\x0b\x32G.inference.ModelInferRequest.InferRequestedOutputTensor.ParametersEntry\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01"\xd5\x04\n\x12ModelInferResponse\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x15\n\rmodel_version\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\x12\x41\n\nparameters\x18\x04 \x03(\x0b\x32-.inference.ModelInferResponse.ParametersEntry\x12@\n\x07outputs\x18\x05 \x03(\x0b\x32/.inference.ModelInferResponse.InferOutputTensor\x12\x1b\n\x13raw_output_contents\x18\x06 \x03(\x0c\x1a\x97\x02\n\x11InferOutputTensor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61tatype\x18\x02 \x01(\t\x12\r\n\x05shape\x18\x03 \x03(\x03\x12S\n\nparameters\x18\x04 \x03(\x0b\x32?.inference.ModelInferResponse.InferOutputTensor.ParametersEntry\x12\x30\n\x08\x63ontents\x18\x05 \x01(\x0b\x32\x1e.inference.InferTensorContents\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01\x1aL\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.inference.InferParameter:\x02\x38\x01"i\n\x0eInferParameter\x12\x14\n\nbool_param\x18\x01 \x01(\x08H\x00\x12\x15\n\x0bint64_param\x18\x02 \x01(\x03H\x00\x12\x16\n\x0cstring_param\x18\x03 \x01(\tH\x00\x42\x12\n\x10parameter_choice"\xd0\x01\n\x13InferTensorContents\x12\x15\n\rbool_contents\x18\x01 \x03(\x08\x12\x14\n\x0cint_contents\x18\x02 \x03(\x05\x12\x16\n\x0eint64_contents\x18\x03 \x03(\x03\x12\x15\n\ruint_contents\x18\x04 \x03(\r\x12\x17\n\x0fuint64_contents\x18\x05 \x03(\x04\x12\x15\n\rfp32_contents\x18\x06 \x03(\x02\x12\x15\n\rfp64_contents\x18\x07 \x03(\x01\x12\x16\n\x0e\x62ytes_contents\x18\x08 \x03(\x0c"0\n\x1aRepositoryModelLoadRequest\x12\x12\n\nmodel_name\x18\x01 \x01(\t"C\n\x1bRepositoryModelLoadResponse\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x10\n\x08isLoaded\x18\x02 \x01(\x08"2\n\x1cRepositoryModelUnloadRequest\x12\x12\n\nmodel_name\x18\x01 \x01(\t"G\n\x1dRepositoryModelUnloadResponse\x12\x12\n\nmodel_name\x18\x01 \x01(\t\x12\x12\n\nisUnloaded\x18\x02 \x01(\x08\x32\xd2\x05\n\x14GRPCInferenceService\x12K\n\nServerLive\x12\x1c.inference.ServerLiveRequest\x1a\x1d.inference.ServerLiveResponse"\x00\x12N\n\x0bServerReady\x12\x1d.inference.ServerReadyRequest\x1a\x1e.inference.ServerReadyResponse"\x00\x12K\n\nModelReady\x12\x1c.inference.ModelReadyRequest\x1a\x1d.inference.ModelReadyResponse"\x00\x12W\n\x0eServerMetadata\x12 .inference.ServerMetadataRequest\x1a!.inference.ServerMetadataResponse"\x00\x12T\n\rModelMetadata\x12\x1f.inference.ModelMetadataRequest\x1a .inference.ModelMetadataResponse"\x00\x12K\n\nModelInfer\x12\x1c.inference.ModelInferRequest\x1a\x1d.inference.ModelInferResponse"\x00\x12\x66\n\x13RepositoryModelLoad\x12%.inference.RepositoryModelLoadRequest\x1a&.inference.RepositoryModelLoadResponse"\x00\x12l\n\x15RepositoryModelUnload\x12\'.inference.RepositoryModelUnloadRequest\x1a(.inference.RepositoryModelUnloadResponse"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "grpc_predict_v2_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY"]._loaded_options = (
        None
    )
    _globals[
        "_MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY"
    ]._serialized_options = b"8\001"
    _globals[
        "_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY"
    ]._loaded_options = None
    _globals[
        "_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY"
    ]._serialized_options = b"8\001"
    _globals["_MODELINFERREQUEST_PARAMETERSENTRY"]._loaded_options = None
    _globals["_MODELINFERREQUEST_PARAMETERSENTRY"]._serialized_options = b"8\001"
    _globals[
        "_MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY"
    ]._loaded_options = None
    _globals[
        "_MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY"
    ]._serialized_options = b"8\001"
    _globals["_MODELINFERRESPONSE_PARAMETERSENTRY"]._loaded_options = None
    _globals["_MODELINFERRESPONSE_PARAMETERSENTRY"]._serialized_options = b"8\001"
    _globals["_SERVERLIVEREQUEST"]._serialized_start = 36
    _globals["_SERVERLIVEREQUEST"]._serialized_end = 55
    _globals["_SERVERLIVERESPONSE"]._serialized_start = 57
    _globals["_SERVERLIVERESPONSE"]._serialized_end = 91
    _globals["_SERVERREADYREQUEST"]._serialized_start = 93
    _globals["_SERVERREADYREQUEST"]._serialized_end = 113
    _globals["_SERVERREADYRESPONSE"]._serialized_start = 115
    _globals["_SERVERREADYRESPONSE"]._serialized_end = 151
    _globals["_MODELREADYREQUEST"]._serialized_start = 153
    _globals["_MODELREADYREQUEST"]._serialized_end = 203
    _globals["_MODELREADYRESPONSE"]._serialized_start = 205
    _globals["_MODELREADYRESPONSE"]._serialized_end = 240
    _globals["_SERVERMETADATAREQUEST"]._serialized_start = 242
    _globals["_SERVERMETADATAREQUEST"]._serialized_end = 265
    _globals["_SERVERMETADATARESPONSE"]._serialized_start = 267
    _globals["_SERVERMETADATARESPONSE"]._serialized_end = 342
    _globals["_MODELMETADATAREQUEST"]._serialized_start = 344
    _globals["_MODELMETADATAREQUEST"]._serialized_end = 397
    _globals["_MODELMETADATARESPONSE"]._serialized_start = 400
    _globals["_MODELMETADATARESPONSE"]._serialized_end = 669
    _globals["_MODELMETADATARESPONSE_TENSORMETADATA"]._serialized_start = 606
    _globals["_MODELMETADATARESPONSE_TENSORMETADATA"]._serialized_end = 669
    _globals["_MODELINFERREQUEST"]._serialized_start = 672
    _globals["_MODELINFERREQUEST"]._serialized_end = 1550
    _globals["_MODELINFERREQUEST_INFERINPUTTENSOR"]._serialized_start = 980
    _globals["_MODELINFERREQUEST_INFERINPUTTENSOR"]._serialized_end = 1256
    _globals[
        "_MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY"
    ]._serialized_start = 1180
    _globals["_MODELINFERREQUEST_INFERINPUTTENSOR_PARAMETERSENTRY"]._serialized_end = (
        1256
    )
    _globals["_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR"]._serialized_start = 1259
    _globals["_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR"]._serialized_end = 1472
    _globals[
        "_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY"
    ]._serialized_start = 1180
    _globals[
        "_MODELINFERREQUEST_INFERREQUESTEDOUTPUTTENSOR_PARAMETERSENTRY"
    ]._serialized_end = 1256
    _globals["_MODELINFERREQUEST_PARAMETERSENTRY"]._serialized_start = 1180
    _globals["_MODELINFERREQUEST_PARAMETERSENTRY"]._serialized_end = 1256
    _globals["_MODELINFERRESPONSE"]._serialized_start = 1553
    _globals["_MODELINFERRESPONSE"]._serialized_end = 2150
    _globals["_MODELINFERRESPONSE_INFEROUTPUTTENSOR"]._serialized_start = 1793
    _globals["_MODELINFERRESPONSE_INFEROUTPUTTENSOR"]._serialized_end = 2072
    _globals[
        "_MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY"
    ]._serialized_start = 1180
    _globals[
        "_MODELINFERRESPONSE_INFEROUTPUTTENSOR_PARAMETERSENTRY"
    ]._serialized_end = 1256
    _globals["_MODELINFERRESPONSE_PARAMETERSENTRY"]._serialized_start = 1180
    _globals["_MODELINFERRESPONSE_PARAMETERSENTRY"]._serialized_end = 1256
    _globals["_INFERPARAMETER"]._serialized_start = 2152
    _globals["_INFERPARAMETER"]._serialized_end = 2257
    _globals["_INFERTENSORCONTENTS"]._serialized_start = 2260
    _globals["_INFERTENSORCONTENTS"]._serialized_end = 2468
    _globals["_REPOSITORYMODELLOADREQUEST"]._serialized_start = 2470
    _globals["_REPOSITORYMODELLOADREQUEST"]._serialized_end = 2518
    _globals["_REPOSITORYMODELLOADRESPONSE"]._serialized_start = 2520
    _globals["_REPOSITORYMODELLOADRESPONSE"]._serialized_end = 2587
    _globals["_REPOSITORYMODELUNLOADREQUEST"]._serialized_start = 2589
    _globals["_REPOSITORYMODELUNLOADREQUEST"]._serialized_end = 2639
    _globals["_REPOSITORYMODELUNLOADRESPONSE"]._serialized_start = 2641
    _globals["_REPOSITORYMODELUNLOADRESPONSE"]._serialized_end = 2712
    _globals["_GRPCINFERENCESERVICE"]._serialized_start = 2715
    _globals["_GRPCINFERENCESERVICE"]._serialized_end = 3437
# @@protoc_insertion_point(module_scope)



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from kserve.configuration import Configuration


class V1alpha1InfereceGraphRouterTimeouts(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'server_idle': 'int',
        'server_read': 'int',
        'server_write': 'int',
        'service_client': 'int'
    }

    attribute_map = {
        'server_idle': 'serverIdle',
        'server_read': 'serverRead',
        'server_write': 'serverWrite',
        'service_client': 'serviceClient'
    }

    def __init__(self, server_idle=None, server_read=None, server_write=None, service_client=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha1InfereceGraphRouterTimeouts - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._server_idle = None
        self._server_read = None
        self._server_write = None
        self._service_client = None
        self.discriminator = None

        if server_idle is not None:
            self.server_idle = server_idle
        if server_read is not None:
            self.server_read = server_read
        if server_write is not None:
            self.server_write = server_write
        if service_client is not None:
            self.service_client = service_client

    @property
    def server_idle(self):
        """Gets the server_idle of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501

        ServerIdle specifies the maximum amount of time in seconds to wait for the next request when keep-alives are enabled.  # noqa: E501

        :return: The server_idle of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501
        :rtype: int
        """
        return self._server_idle

    @server_idle.setter
    def server_idle(self, server_idle):
        """Sets the server_idle of this V1alpha1InfereceGraphRouterTimeouts.

        ServerIdle specifies the maximum amount of time in seconds to wait for the next request when keep-alives are enabled.  # noqa: E501

        :param server_idle: The server_idle of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501
        :type: int
        """

        self._server_idle = server_idle

    @property
    def server_read(self):
        """Gets the server_read of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501

        ServerRead specifies the number of seconds to wait before timing out a request read by the server.  # noqa: E501

        :return: The server_read of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501
        :rtype: int
        """
        return self._server_read

    @server_read.setter
    def server_read(self, server_read):
        """Sets the server_read of this V1alpha1InfereceGraphRouterTimeouts.

        ServerRead specifies the number of seconds to wait before timing out a request read by the server.  # noqa: E501

        :param server_read: The server_read of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501
        :type: int
        """

        self._server_read = server_read

    @property
    def server_write(self):
        """Gets the server_write of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501

        ServerWrite specifies the maximum duration in seconds before timing out writes of the response.  # noqa: E501

        :return: The server_write of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501
        :rtype: int
        """
        return self._server_write

    @server_write.setter
    def server_write(self, server_write):
        """Sets the server_write of this V1alpha1InfereceGraphRouterTimeouts.

        ServerWrite specifies the maximum duration in seconds before timing out writes of the response.  # noqa: E501

        :param server_write: The server_write of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501
        :type: int
        """

        self._server_write = server_write

    @property
    def service_client(self):
        """Gets the service_client of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501

        ServiceClient specifies a time limit in seconds for requests made to the graph components by HTTP client.  # noqa: E501

        :return: The service_client of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501
        :rtype: int
        """
        return self._service_client

    @service_client.setter
    def service_client(self, service_client):
        """Sets the service_client of this V1alpha1InfereceGraphRouterTimeouts.

        ServiceClient specifies a time limit in seconds for requests made to the graph components by HTTP client.  # noqa: E501

        :param service_client: The service_client of this V1alpha1InfereceGraphRouterTimeouts.  # noqa: E501
        :type: int
        """

        self._service_client = service_client

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha1InfereceGraphRouterTimeouts):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha1InfereceGraphRouterTimeouts):
            return True

        return self.to_dict() != other.to_dict()



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
KServe

Python SDK for KServe  # noqa: E501

The version of the OpenAPI document: v0.1
Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import datetime
import unittest

from kserve.models.v1alpha1_local_model_node_group_list import (  # noqa: E501
    V1alpha1LocalModelNodeGroupList,
)
from kserve.rest import ApiException

import kserve


class TestV1alpha1LocalModelNodeGroupList(unittest.TestCase):
    """V1alpha1LocalModelNodeGroupList unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1alpha1LocalModelNodeGroupList
        include_option is a boolean, when False only required
        params are included, when True both required and
        optional params are included"""
        # model = kserve.models.v1alpha1_local_model_node_group_list.V1alpha1LocalModelNodeGroupList()  # noqa: E501
        if include_optional:
            return V1alpha1LocalModelNodeGroupList(
                api_version="0",
                items=[
                    kserve.models.v1alpha1_local_model_node_group.V1alpha1LocalModelNodeGroup(
                        api_version="0",
                        kind="0",
                        metadata=None,
                        spec=kserve.models.v1alpha1_local_model_node_group_spec.V1alpha1LocalModelNodeGroupSpec(
                            storage_limit="1Gi",
                            persistent_volume_claim_spec="",
                            persistent_volume_spec="",
                        ),
                    )
                ],
                kind="0",
                metadata=None,
            )
        else:
            return V1alpha1LocalModelNodeGroupList(
                items=[
                    kserve.models.v1alpha1_local_model_node_group.V1alpha1LocalModelNodeGroup(
                        api_version="0",
                        kind="0",
                        metadata=None,
                        spec=kserve.models.v1alpha1_local_model_node_group_spec.V1alpha1LocalModelNodeGroupSpec(
                            storage_limit="1Gi",
                            persistent_volume_claim_spec="",
                            persistent_volume_spec="",
                        ),
                    )
                ],
            )

    def testV1alpha1LocalModelNodeGroupList(self):
        """Test V1alpha1LocalModelNodeGroupList"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
KServe

Python SDK for KServe  # noqa: E501

The version of the OpenAPI document: v0.1
Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1beta1_inference_service import (
    V1beta1InferenceService,
)  # noqa: E501
from kserve.rest import ApiException


class TestV1beta1InferenceService(unittest.TestCase):
    """V1beta1InferenceService unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1beta1InferenceService
        include_option is a boolean, when False only required
        params are included, when True both required and
        optional params are included"""
        # model = kserve.models.v1beta1_inference_service.V1beta1InferenceService()  # noqa: E501
        if include_optional:
            return V1beta1InferenceService(
                api_version="0",
                kind="0",
                metadata=None,
                spec=kserve.models.v1beta1_inference_service_spec.V1beta1InferenceServiceSpec(
                    explainer=kserve.models.v1beta1_explainer_spec.V1beta1ExplainerSpec(
                        active_deadline_seconds=56,
                        affinity=None,
                        art=kserve.models.v1beta1_art_explainer_spec.V1beta1ARTExplainerSpec(
                            image="0",
                            image_pull_policy="0",
                            lifecycle=None,
                            liveness_probe=None,
                            name="0",
                            readiness_probe=None,
                            resources=None,
                            runtime_version="0",
                            security_context=None,
                            startup_probe=None,
                            stdin=True,
                            stdin_once=True,
                            storage_uri="0",
                            termination_message_path="0",
                            termination_message_policy="0",
                            tty=True,
                            type="0",
                            working_dir="0",
                        ),
                        automount_service_account_token=True,
                        batcher=kserve.models.v1beta1_batcher.V1beta1Batcher(
                            max_batch_size=56,
                            max_latency=56,
                            timeout=56,
                        ),
                        canary_traffic_percent=56,
                        container_concurrency=56,
                        containers=[None],
                        dns_config=None,
                        dns_policy="0",
                        enable_service_links=True,
                        ephemeral_containers=[None],
                        host_aliases=[None],
                        host_ipc=True,
                        host_network=True,
                        host_pid=True,
                        hostname="0",
                        image_pull_secrets=[None],
                        init_containers=[None],
                        logger=kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                            mode="0",
                            url="0",
                        ),
                        max_replicas=56,
                        min_replicas=56,
                        node_name="0",
                        node_selector={"key": "0"},
                        overhead={"key": None},
                        preemption_policy="0",
                        priority=56,
                        priority_class_name="0",
                        readiness_gates=[None],
                        restart_policy="0",
                        runtime_class_name="0",
                        scheduler_name="0",
                        security_context=None,
                        service_account="0",
                        service_account_name="0",
                        share_process_namespace=True,
                        subdomain="0",
                        termination_grace_period_seconds=56,
                        timeout=56,
                        tolerations=[None],
                        topology_spread_constraints=[None],
                        volumes=[None],
                    ),
                    predictor=kserve.models.v1beta1_predictor_spec.V1beta1PredictorSpec(
                        active_deadline_seconds=56,
                        affinity=None,
                        automount_service_account_token=True,
                        canary_traffic_percent=56,
                        container_concurrency=56,
                        dns_config=None,
                        dns_policy="0",
                        enable_service_links=True,
                        host_ipc=True,
                        host_network=True,
                        host_pid=True,
                        hostname="0",
                        max_replicas=56,
                        min_replicas=56,
                        node_name="0",
                        onnx=kserve.models.v1beta1_onnx_runtime_spec.V1beta1ONNXRuntimeSpec(
                            image="0",
                            image_pull_policy="0",
                            lifecycle=None,
                            liveness_probe=None,
                            name="0",
                            readiness_probe=None,
                            resources=None,
                            runtime_version="0",
                            security_context=None,
                            startup_probe=None,
                            stdin=True,
                            stdin_once=True,
                            storage_uri="0",
                            termination_message_path="0",
                            termination_message_policy="0",
                            tty=True,
                            working_dir="0",
                        ),
                        preemption_policy="0",
                        priority=56,
                        priority_class_name="0",
                        pytorch=kserve.models.v1beta1_torch_serve_spec.V1beta1TorchServeSpec(
                            image="0",
                            image_pull_policy="0",
                            lifecycle=None,
                            liveness_probe=None,
                            name="0",
                            readiness_probe=None,
                            resources=None,
                            runtime_version="0",
                            security_context=None,
                            startup_probe=None,
                            stdin=True,
                            stdin_once=True,
                            storage_uri="0",
                            termination_message_path="0",
                            termination_message_policy="0",
                            tty=True,
                            working_dir="0",
                        ),
                        restart_policy="0",
                        runtime_class_name="0",
                        scheduler_name="0",
                        security_context=None,
                        service_account="0",
                        service_account_name="0",
                        share_process_namespace=True,
                        sklearn=kserve.models.v1beta1_sk_learn_spec.V1beta1SKLearnSpec(
                            image="0",
                            image_pull_policy="0",
                            lifecycle=None,
                            liveness_probe=None,
                            name="0",
                            readiness_probe=None,
                            resources=None,
                            runtime_version="0",
                            security_context=None,
                            startup_probe=None,
                            stdin=True,
                            stdin_once=True,
                            storage_uri="0",
                            termination_message_path="0",
                            termination_message_policy="0",
                            tty=True,
                            working_dir="0",
                        ),
                        subdomain="0",
                        tensorflow=kserve.models.v1beta1_tf_serving_spec.V1beta1TFServingSpec(
                            image="0",
                            image_pull_policy="0",
                            lifecycle=None,
                            liveness_probe=None,
                            name="0",
                            readiness_probe=None,
                            resources=None,
                            runtime_version="0",
                            security_context=None,
                            startup_probe=None,
                            stdin=True,
                            stdin_once=True,
                            storage_uri="0",
                            termination_message_path="0",
                            termination_message_policy="0",
                            tty=True,
                            working_dir="0",
                        ),
                        termination_grace_period_seconds=56,
                        timeout=56,
                        triton=kserve.models.v1beta1_triton_spec.V1beta1TritonSpec(
                            image="0",
                            image_pull_policy="0",
                            lifecycle=None,
                            liveness_probe=None,
                            name="0",
                            readiness_probe=None,
                            resources=None,
                            runtime_version="0",
                            security_context=None,
                            startup_probe=None,
                            stdin=True,
                            stdin_once=True,
                            storage_uri="0",
                            termination_message_path="0",
                            termination_message_policy="0",
                            tty=True,
                            working_dir="0",
                        ),
                        xgboost=kserve.models.v1beta1_xg_boost_spec.V1beta1XGBoostSpec(
                            image="0",
                            image_pull_policy="0",
                            lifecycle=None,
                            liveness_probe=None,
                            name="0",
                            readiness_probe=None,
                            resources=None,
                            runtime_version="0",
                            security_context=None,
                            startup_probe=None,
                            stdin=True,
                            stdin_once=True,
                            storage_uri="0",
                            termination_message_path="0",
                            termination_message_policy="0",
                            tty=True,
                            working_dir="0",
                        ),
                    ),
                    transformer=kserve.models.v1beta1_transformer_spec.V1beta1TransformerSpec(
                        active_deadline_seconds=56,
                        affinity=None,
                        automount_service_account_token=True,
                        canary_traffic_percent=56,
                        container_concurrency=56,
                        dns_config=None,
                        dns_policy="0",
                        enable_service_links=True,
                        host_ipc=True,
                        host_network=True,
                        host_pid=True,
                        hostname="0",
                        max_replicas=56,
                        min_replicas=56,
                        node_name="0",
                        preemption_policy="0",
                        priority=56,
                        priority_class_name="0",
                        restart_policy="0",
                        runtime_class_name="0",
                        scheduler_name="0",
                        security_context=None,
                        service_account="0",
                        service_account_name="0",
                        share_process_namespace=True,
                        subdomain="0",
                        termination_grace_period_seconds=56,
                        timeout=56,
                    ),
                ),
                status=kserve.models.v1beta1_inference_service_status.V1beta1InferenceServiceStatus(
                    address=None,
                    components={
                        "key": kserve.models.v1beta1_component_status_spec.V1beta1ComponentStatusSpec(
                            address=None,
                            latest_created_revision="0",
                            latest_ready_revision="0",
                            latest_rolledout_revision="0",
                            previous_rolledout_revision="0",
                            traffic=list(),
                            url=None,
                        )
                    },
                    conditions=[None],
                    observed_generation=56,
                    url=None,
                ),
            )
        else:
            return V1beta1InferenceService()

    def testV1beta1InferenceService(self):
        """Test V1beta1InferenceService"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == "__main__":
    unittest.main()
